"""Test cascade_config."""

import argparse
import json
from os import stat
import tempfile
from typing import Type

import pytest
import jsonschema

import cascade_config
from cascade_config import ValidationSchema

TEST_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Test schema",
    "type": "object",
    "properties": {
        "config_example": {
            "type": "object",
            "properties": {
                "num_cpu": {
                    "type": "number",
                    "minimum": -1,
                },
                "log_level": {
                    "type": "string",
                    "enum": ["debug", "info", "warning", "error", "critical"]
                }
            }
        }
    }
}

TEST_SAMPLE = {
    "config_example": {"num_cpu": 1, "log_level": "info"}
}
TEST_SAMPLE_2 = {
    "config_example": {"log_level": "debug"}, "test": True
}
TEST_SAMPLE_CASC = {
    "config_example": {"num_cpu": 1, "log_level": "debug"}, "test": True
}

TEST_SAMPLE_INVALID = {
    "config_example": {"num_cpu": "not_a_number", "log_level": "info"}
}

def get_sample_namespace(test_sample):
    flatten = lambda l: [item for sublist in l for item in sublist]
    test_args = flatten([[f"--{i[0]}", f"{i[1]}"] for i in test_sample.items()])
    parser = argparse.ArgumentParser()
    parser.add_argument("--num_cpu", type=int)
    parser.add_argument("--log_level")
    return parser.parse_args(test_args)


class TestCascadeConfig:
    """Test CascadeConfig."""

    @staticmethod
    def get_json_file(test_sample):
        with tempfile.NamedTemporaryFile(mode='wt', delete=False) as json_file:
            json.dump(test_sample, json_file)
            json_file.seek(0)
            json_file_name = json_file.name
        return json_file_name

    @staticmethod
    def get_sample_namespace(test_sample):
        flatten = lambda l: [item for sublist in l for item in sublist]
        test_args = flatten([[f"--{i[0]}", f"{i[1]}"] for i in test_sample.items()])
        parser = argparse.ArgumentParser()
        parser.add_argument("--num_cpu", type=int)
        parser.add_argument("--log_level")
        return parser.parse_args(test_args)

    def test_single_config(self):
        cc = cascade_config.CascadeConfig()
        cc.add_dict(TEST_SAMPLE)
        assert cc.parse() == TEST_SAMPLE

    def test_single_config_validation(self):
        """Test single valid dict source, with JSON Schema validation"."""
        cc = cascade_config.CascadeConfig(validation_schema=TEST_SCHEMA)
        cc.add_dict(TEST_SAMPLE)
        assert cc.parse() == TEST_SAMPLE

    def test_single_config_validation_invalid(self):
        """Test single invalid dict source, with JSON Schema validation"."""
        cc = cascade_config.CascadeConfig(validation_schema=TEST_SCHEMA)
        cc.add_dict(TEST_SAMPLE_INVALID)
        with pytest.raises(jsonschema.exceptions.ValidationError):
            cc.parse()

    def test_single_config_json(self):
        """Test JSON config."""
        cc = cascade_config.CascadeConfig()
        cc.add_json(self.get_json_file(TEST_SAMPLE), validation_schema=TEST_SCHEMA)
        assert cc.parse() == TEST_SAMPLE

    def test_single_config_namespace(self):
        """Test Namespace config."""
        subkey = "config_example"
        cc = cascade_config.CascadeConfig()
        cc.add_namespace(
            get_sample_namespace(TEST_SAMPLE[subkey]),
            subkey=subkey,
            validation_schema=TEST_SCHEMA
        )
        assert cc.parse() == TEST_SAMPLE

    def test_single_config_dict_typeerror(self):
        """Test Namespace config type error."""
        cc = cascade_config.CascadeConfig()
        cc.add_dict(42)
        with pytest.raises(TypeError):
            cc.parse()

    def test_single_config_json_typeerror(self):
        """Test Namespace config type error."""
        cc = cascade_config.CascadeConfig()
        cc.add_json(42)
        with pytest.raises(TypeError):
            cc.parse()

    def test_single_config_namespace_typeerror(self):
        """Test Namespace config type error."""
        cc = cascade_config.CascadeConfig()
        cc.add_namespace("not_a_namespace")
        with pytest.raises(TypeError):
            cc.parse()

    def test_multiple_configs(self):
        """Test multiple config sources."""
        subkey = "config_example"
        cc = cascade_config.CascadeConfig(validation_schema=TEST_SCHEMA)
        cc.add_namespace(
            get_sample_namespace(TEST_SAMPLE[subkey]),
            subkey=subkey,
        )
        cc.add_json(self.get_json_file(TEST_SAMPLE_2))
        assert cc.parse() == TEST_SAMPLE_CASC

    def test_validation_schema_from_object(self):
        with pytest.raises(TypeError):
            cascade_config.ValidationSchema.from_object(42)

    def test_validation_schema_from_json(self):
        vs = cascade_config.ValidationSchema(self.get_json_file(TEST_SCHEMA))
        assert vs.load() == TEST_SCHEMA



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from m2r import MdInclude
from recommonmark.transform import AutoStructify

sys.path.insert(0, os.path.abspath('../../'))

# -- Project information -----------------------------------------------------

project = 'cascade-config'
copyright = '2020, RalfG'
author = 'RalfG'
github_project_url = "https://github.com/ralfg/cascade-config/"
github_doc_root = 'https://github.com/ralfg/cascade-config/tree/master/docs/'


def get_version(path):
    with open(path, "rt") as f:
        for line in f:
            if line.startswith("__version__ = "):
                return line.strip().split(" = ")[1].strip("\"'")

# The full version, including alpha/beta/rc tags
release = get_version("../../cascade_config.py")

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx_rtd_theme",
    "recommonmark",
]
source_suffix = ['.rst', '.md']
master_doc = 'index'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Autodoc options --------------------------------------------------------
autodoc_member_order = 'bysource'
autoclass_content = 'init'


# https://github.com/rtfd/recommonmark/blob/master/docs/conf.py
def setup(app):
    config = {
        # 'url_resolver': lambda url: github_doc_root + url,
        'auto_toc_tree_section': 'Contents',
        'enable_eval_rst': True,
    }
    app.add_config_value('recommonmark_config', config, True)
    app.add_transform(AutoStructify)

    # from m2r to make `mdinclude` work
    app.add_config_value('no_underscore_emphasis', False, 'env')
    app.add_config_value('m2r_parse_relative_links', False, 'env')
    app.add_config_value('m2r_anonymous_references', False, 'env')
    app.add_config_value('m2r_disable_inline_math', False, 'env')
    app.add_directive('mdinclude', MdInclude)



"""Cascading configuration from the CLI and config files."""

__version__ = "0.3.0"

import json
import os
from abc import ABC, abstractmethod
from argparse import ArgumentParser, Namespace
from typing import Dict

import jsonschema


class CascadeConfig:
    """Cascading configuration."""

    def __init__(self, validation_schema=None, none_overrides_value=False):
        """
        Cascading configuration.

        Parameters
        ----------
        validation_schema: str, path-like, dict, or cascade_config.ValidationSchema, optional
            JSON Schema to validate fully cascaded configuration
        none_overrides_value: bool
            If True, a None value overrides a not-None value from the previous configuration.
            If False, None values will never override not-None values.

        Examples
        --------
        >>> cascade_conf = CascadeConfig(validation_schema="config_schema.json")
        >>> cascade_conf.add_json("config_default.json")
        >>> cascade_conf.add_json("config_user.json")
        >>> config = cascade_conf.parse()

        """
        self.validation_schema = validation_schema
        self.none_overrides_value = none_overrides_value
        self.sources = []

    @property
    def validation_schema(self):
        """JSON Schema to validate fully cascaded configuration."""
        return self._validation_schema

    @validation_schema.setter
    def validation_schema(self, value):
        """Set validation schema."""
        if value:
            self._validation_schema = ValidationSchema.from_object(value)
        else:
            self._validation_schema = None

    def _update_dict_recursively(self, original: Dict, updater: Dict) -> Dict:
        """Update dictionary recursively."""
        for k, v in updater.items():
            if isinstance(v, dict):
                original[k] = self._update_dict_recursively(original.get(k, {}), v)
            elif v or k not in original:  # v is not None, or key does not exist yet
                original[k] = v
            elif self.none_overrides_value:  # v is None, but can override previous value
                original[k] = v
        return original

    def add_dict(self, *args, **kwargs):
        """
        Add dictionary configuration source to source list.
        *args and **kwargs are passed to :class:`cascade_config.DictConfigSource()`.

        """
        source = DictConfigSource(*args, **kwargs)
        self.sources.append(source)

    def add_argumentparser(self, *args, **kwargs):
        """
        Add argumentparser configuration source to source list.
        *args and **kwargs are passed to :class:`cascade_config.ArgumentParserConfigSource()`.

        """
        source = ArgumentParserConfigSource(*args, **kwargs)
        self.sources.append(source)

    def add_namespace(self, *args, **kwargs):
        """
        Add argparse Namespace configuration source to source list.
        *args and **kwargs are passed to :class:`cascade_config.NamespaceConfigSource()`.
        """
        source = NamespaceConfigSource(*args, **kwargs)
        self.sources.append(source)

    def add_json(self, *args, **kwargs):
        """
        Add JSON configuration source to source list.
        *args and **kwargs are passed to :class:`cascade_config.JSONConfigSource()`.
        """
        source = JSONConfigSource(*args, **kwargs)
        self.sources.append(source)

    def parse(self) -> Dict:
        """Parse all sources, cascade, validate, and return cascaded configuration."""
        config = dict()
        for source in self.sources:
            config = self._update_dict_recursively(config, source.load())

        if self.validation_schema:
            jsonschema.validate(config, self.validation_schema.load())

        return config


class _ConfigSource(ABC):
    """Abstract base class for configuration source."""

    def __init__(self, source, validation_schema=None, subkey=None) -> None:
        """
        Initialize a single configuration source.

        Parameters
        ----------
        source : str, path-like, dict, argparse.ArgumentParser
            source for the configuration, either a dictionary, path to a file, or
            argument parser.
        validation_schema: str, path-like, dict, or cascade_config.ValidationSchema, optional
            JSON Schema to validate single configuration
        subkey : str
            adds the configuration to a subkey of the final cascased configuration;
            e.g. specifying a subkey `"user"` for a configuration source, would add it
            under the key `"user"` in the cascaded configuration, instead of updating
            the root of the existing configuration

        Methods
        -------
        load()
            load the configuration from the source and return it as a dictionary

        """
        self.source = source
        self.validation_schema = validation_schema
        self.subkey = subkey

    @property
    def validation_schema(self):
        """Get validation_schema."""
        return self._validation_schema

    @validation_schema.setter
    def validation_schema(self, value):
        """Set validation schema."""
        if value:
            self._validation_schema = ValidationSchema.from_object(value)
        else:
            self._validation_schema = None

    @abstractmethod
    def _read(self):
        """Read source into dict."""
        pass

    def load(self) -> Dict:
        """Read, validate, and place in subkey if required."""
        if self.subkey:
            config = dict()
            config[self.subkey] = self._read()
        else:
            config = self._read()
        if self.validation_schema:
            jsonschema.validate(config, self.validation_schema.load())
        return config


class DictConfigSource(_ConfigSource):
    """Dictionary configuration source."""

    def _read(self) -> Dict:
        if not isinstance(self.source, dict):
            raise TypeError("DictConfigSource `source` must be a dict")
        return self.source


class JSONConfigSource(_ConfigSource):
    """JSON configuration source."""

    def _read(self) -> Dict:
        if not isinstance(self.source, (str, os.PathLike)):
            raise TypeError(
                "JSONConfigSource `source` must be a string or path-like object"
            )
        with open(self.source, "rt") as json_file:
            config = json.load(json_file)
        return config


class ArgumentParserConfigSource(_ConfigSource):
    """ArgumentParser configuration source."""

    def _read(self) -> Dict:
        if not isinstance(self.source, ArgumentParser):
            raise TypeError(
                "ArgumentParserSource `source` must be an argparse.ArgumentParser object"
            )
        config = vars(self.source.parse_args())
        return config


class NamespaceConfigSource(_ConfigSource):
    """Argparse Namespace configuration source."""

    def _read(self) -> Dict:
        if not isinstance(self.source, Namespace):
            raise TypeError(
                "NamespaceConfigSource `source` must be an argparse.Namespace object"
            )
        config = vars(self.source)
        return config


class ValidationSchema:
    """ValidationSchema."""

    def __init__(self, source):
        """ValidationSchema."""
        self.source = source

    @classmethod
    def from_object(cls, obj):
        """Return ValidationSchema from str, path-like, dict, or ValidationSchema."""
        if isinstance(obj, (str, os.PathLike, Dict)):
            return cls(obj)
        elif isinstance(obj, cls):
            return obj
        else:
            raise TypeError(
                f"Cannot create ValidationSchema from type {type(obj)}. Must be a "
                "string, path-like, dict, or cascade_config.ValidationSchema object"
            )

    def load(self) -> Dict:
        """Load validation schema."""
        if isinstance(self.source, (str, os.PathLike)):
            with open(self.source, "rt") as json_file:
                schema = json.load(json_file)
        elif isinstance(self.source, Dict):
            schema = self.source
        else:
            raise TypeError(
                "ValidationSchema `source` must be of type string, path-like, or dict"
            )
        return schema



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

install_requires = \
['jsonschema>=3,<4']

extras_require = \
{'test': ['pytest>=6,<7', 'pytest-cov>=2,<3']}

setup(name='cascade-config',
      version='0.3.0',
      description='Cascading configuration from the CLI and config files.',
      author='RalfG',
      author_email='ralf@gabriels.dev',
      url='https://github.com/RalfG/cascade-config',
      py_modules=['cascade_config'],
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
